header {* A theory of generalization *}

theory Generalization
  imports Main_ZF ZF_Addons
begin

text {*
  This theory formalizes the intuitive notion of \emph{generalization}.
*}

subsection {* Generalization situation *}

locale generalization_situation =
  fixes small::i and big::i
  fixes embed::i
  assumes embed_inj: "embed\<in>inj(small, big)"
begin

text {*
  In Isabelle 2009-2 locale params are not visible.
  As a workaround I define their abbreviations:
*}
abbreviation "small_param \<equiv> small"
abbreviation "big_param \<equiv> big"
abbreviation "embed_param \<equiv> embed"

definition small2_def: "small2 \<equiv> range(embed)"
definition spec_def: "spec \<equiv> converse(embed)"

lemma spec_inj: "spec\<in>inj(small2, small)"
proof -
  from embed_inj have "converse(embed)\<in>inj(range(embed), small)" by (rule inj_converse_inj)
  with small2_def spec_def show ?thesis by simp
qed

lemma spec_fun: "spec: small2\<rightarrow>small"
proof -
  from embed_inj have "converse(embed): range(embed)\<rightarrow>small" by (rule inj_converse_fun)
  with small2_def spec_def show ?thesis by simp
qed

lemma embed_fun: "embed: small\<rightarrow>big" by (rule inj_is_fun[OF embed_inj])

lemma embed_surj: "embed\<in>surj(small, small2)"
proof -
  have "embed\<in>surj(small, range(embed))" by (rule fun_is_surj[OF embed_fun])
  hence ?thesis "embed\<in>surj(small, small2)" by (unfold small2_def)
  show ?thesis by fact
qed

theorem embed_bij: "embed\<in>bij(small, small2)"
proof -
  from embed_inj have "embed\<in>bij(small, range(embed))" using inj_bij_range by auto
  thus ?thesis by (unfold small2_def)
qed

theorem small2_sub_big: "small2\<subseteq>big" using embed_fun range_subset by (auto simp add: small2_def)

theorem spec_bij: "spec\<in>bij(small2, small)"
proof -
  have "converse(embed)\<in>bij(small2, small)" by (rule bij_converse_bij[OF embed_bij])
  thus ?thesis by (unfold spec_def)
qed

end

subsection {* Arbitrary generalizations *}

locale arbitrary_generalization = generalization_situation +
  fixes newbig::i
  fixes move::i
  assumes
  move_bij: "move\<in>bij(big, newbig)" and
  move_embed: "move O embed = id(small)"
begin

text {*
  In Isabelle 2009-2 locale params are not visible.
  As a workaround I define their abbreviations:
*}
abbreviation "newbig_param \<equiv> newbig"
abbreviation "move_param \<equiv> move"

lemma move_fun: "move: big\<rightarrow>newbig" using move_bij bij_is_fun by auto

lemma move_inj: "move\<in>inj(big, newbig)" using move_bij by (rule bij_is_inj)
lemma move_surj: "move\<in>surj(big, newbig)" using move_bij by (rule bij_is_surj)

lemma move_domain: "domain(move) = big" using domain_of_fun [OF move_fun] by auto

theorem move_embed_plain [simp]: "x\<in>small \<Longrightarrow> move`(embed`x) = x"
proof -
  assume "x: small"
  with embed_fun have "move`(embed`x) = (move O embed)`x" using comp_fun_apply by auto
  also have "... = id(small)`x" by (simp add: move_embed)
  also with `x\<in>small` have "id(small)`x = x" by simp
  finally show "move`(embed`x) = x" by simp
qed

definition ret_def: "ret \<equiv> converse(move)"

lemma ret_bij: "ret\<in>bij(newbig, big)" using move_bij unfolding ret_def by auto

lemma ret_inj: "ret\<in>inj(newbig, big)" using ret_bij by (rule bij_is_inj)
lemma ret_surj: "ret\<in>surj(newbig, big)" using ret_bij by (rule bij_is_surj)

lemma ret_restrict: "embed = restrict(ret, small)"
proof -
  have "converse(move) O move O embed = converse(move) O id(small)" using move_embed by auto
  hence "(converse(move) O move) O embed = converse(move) O id(small)" using comp_assoc by auto
  moreover
  with left_comp_inverse move_inj have i [simp]: "converse(move) O move = id(big)" by simp
  ultimately have a: "id(big) O embed = converse(move) O id(small)" by simp
  with embed_fun have "embed\<subseteq>small\<times>big" using fun_is_rel by auto
  hence "id(big) O embed = embed" using left_comp_id by auto
  with a have "embed = converse(move) O id(small)" by auto
  hence "embed = restrict(converse(move), small)" using right_comp_id_any by auto
  thus "embed = restrict(ret, small)" using right_comp_id_any unfolding ret_def by auto
qed

end

subsection {* ZF generalization *}

text {* We will need this lemma to assert that ZF generalization is an arbitrary generalization: *}

lemma mem_not_refl_2: "{t}\<notin>t"
proof
  assume as: "{t}\<in>t"
  let ?A = "{t,{t}}"
  have nz: "?A \<noteq> 0" by auto
  have "\<forall>x\<in>?A. \<exists>y\<in>x. y\<in>?A"
  proof
    fix "x" assume "x\<in>?A"
    have x_in_A: "x\<in>?A" by fact
    have ex: "\<exists>y\<in>x. y\<in>?A"
    proof cases
      assume "x=t"
      with as have "{t}\<in>x \<and> {t}\<in>?A" by auto
      thus ?thesis by auto
    next
      assume "x\<noteq>t"
      with x_in_A have "x={t}" by auto
      with as have "t\<in>x \<and> t\<in>?A" by auto
      thus ?thesis by auto
    qed
    thus "\<exists>y\<in>x. y\<in>?A" by auto
  qed
  with foundation show False by auto
qed

locale ZF_generalization = generalization_situation
begin

definition token_def: "token \<equiv> Pow(\<Union>(\<Union>(small)))"

lemma token_not_small: "\<langle>token,x\<rangle>\<notin>small"
proof
  assume "\<langle>token,x\<rangle>\<in>small"
  hence "{{token,token}, {token,x}}\<in>small" by (simp add: Pair_def)
  hence "{{token}, {token,x}}\<in>small" by auto
  hence "{token}\<in> \<Union>(small)" by auto
  hence "{token} \<subseteq> \<Union>(\<Union>(small))" by auto
  hence "{token}\<in>Pow(\<Union>(\<Union>(small)))" by auto
  hence "{token}\<in>token" by (simp add: token_def)
  with mem_not_refl_2 show False by contradiction
qed

definition zf_move_fun::"i\<Rightarrow>i" where zf_move_fun_def: "zf_move_fun(x) \<equiv> if x: small2 then spec`x else \<langle>token,x\<rangle>"
definition zf_move_def: "zf_move \<equiv> (lam x:big. zf_move_fun(x))"

definition zf_newbig_def: "zf_newbig \<equiv> range(zf_move)"

lemma zf_move_domain: "domain(zf_move) = big" using domain_lam by (auto simp add: zf_move_def)

lemma zf_move_fun: "zf_move: big\<rightarrow>zf_newbig" by (simp add: lam_is_fun zf_move_def zf_newbig_def)

theorem small_less_zf_newbig: "small \<subseteq> zf_newbig"
proof
  fix x
  assume s: "x\<in>small"
  with embed_fun have "embed`x\<in>range(embed)" by (rule apply_rangeI)
  hence s1: "embed`x\<in>small2" by (unfold small2_def)
  with small2_sub_big have s2: "embed`x\<in>big" by auto
  hence "zf_move`(embed`x) = zf_move_fun(embed`x)" by (auto simp add: zf_move_def)
  with s1 have "zf_move`(embed`x) = spec`(embed`x)" by (simp add: zf_move_fun_def)
  with embed_inj s spec_def have x_val: "zf_move`(embed`x) = x" using left_inverse by auto
  from zf_move_fun s2 have "zf_move`(embed`x)\<in>range(zf_move)" by (rule apply_rangeI)
  with x_val have "x\<in>range(zf_move)" by auto
  with x_val zf_newbig_def show "x\<in>zf_newbig" by auto
qed

theorem zf_move_inj: "zf_move\<in>inj(big, zf_newbig)"
proof -
  have "\<forall>a\<in>big. \<forall>b\<in>big. zf_move`a = zf_move`b \<longrightarrow> a=b"
  proof -
    {
      fix "a" "b"
      assume "a\<in>big" "b\<in>big"
      assume move_eq: "zf_move`a = zf_move`b"
      have spec1_a: "a:small2 \<Longrightarrow> zf_move`a = spec`a"
      proof -
        assume "a\<in>small2"
        with `a\<in>big` show ?thesis by (simp add: zf_move_def zf_move_fun_def)
      qed
      have spec1_b: "b\<in>small2 \<Longrightarrow> zf_move`b = spec`b"
      proof -
        assume "b\<in>small2"
        with `b\<in>big` show ?thesis by (simp add: zf_move_def zf_move_fun_def)
      qed
      have spec2_a: "a\<notin>small2 \<Longrightarrow> zf_move`a = \<langle>token,a\<rangle>"
      proof -
        assume "a\<notin>small2"
        with `a\<in>big` show ?thesis by (simp add: zf_move_def zf_move_fun_def)
      qed
      have spec2_b: "b\<notin>small2 \<Longrightarrow> zf_move`b = \<langle>token,b\<rangle>"
      proof -
        assume "b\<notin>small2"
        with `b\<in>big` show ?thesis by (simp add: zf_move_def zf_move_fun_def)
      qed
      have "a=b"
      proof -
        { assume "a\<in>small2" "b\<in>small2"
          from `a\<in>small2` have "zf_move`a = spec`a" by (rule spec1_a)
          moreover
          from `b\<in>small2` have "zf_move`b = spec`b" by (rule spec1_b)
          ultimately
          have "spec`a = spec`b" using move_eq by auto
          with spec_inj `a:small2` `b:small2` have "a=b" using inj_def by auto
        }
        moreover
        { assume "a\<in>small2" "b\<notin>small2"
          from `a\<in>small2` have "zf_move`a = spec`a" by (rule spec1_a)
          with spec_fun `a\<in>small2` have ma_s: "zf_move`a\<in>small" using apply_funtype by auto
          from `b\<notin>small2` have "zf_move`b = \<langle>token,b\<rangle>" by (rule spec2_b)
          hence "zf_move`b\<notin>small" using token_not_small by auto
          with move_eq ma_s have False by auto
        }
        moreover
        { assume "a\<notin>small2" "b\<in>small2"
          from `b\<in>small2` have "zf_move`b = spec`b" by (rule spec1_b)
          with spec_fun `b\<in>small2` have mb_s: "zf_move`b\<in>small" using apply_funtype by auto
          from `a\<notin>small2` have "zf_move`a = \<langle>token,a\<rangle>" by (rule spec2_a)
          hence "zf_move`a\<notin>small" using token_not_small by auto
          with move_eq mb_s have False by auto
        }
        moreover
        { assume "a\<notin>small2" "b\<notin>small2"
          from `a\<notin>small2` have mt_a: "zf_move`a = \<langle>token,a\<rangle>" by (rule spec2_a)
          moreover
          from `b\<notin>small2` have mt_b: "zf_move`b = \<langle>token,b\<rangle>" by (rule spec2_b)
          from move_eq mt_a mt_b have "\<langle>token,a\<rangle>=\<langle>token,b\<rangle>" by auto -- "Order of from conditions is important to not cause infinite loop"
          hence "a=b" by auto
        }
        ultimately show "a=b" by auto
      qed
    }
    thus ?thesis by auto
  qed
  with zf_move_fun show ?thesis using inj_def by simp
qed

theorem zf_move_surj: "zf_move\<in>surj(big, zf_newbig)"
proof -
  have "zf_move\<in>surj(big, range(zf_move))" by (rule fun_is_surj[OF zf_move_fun])
  hence ?thesis "zf_move\<in>surj(big, zf_newbig)" by (unfold zf_newbig_def)
  show ?thesis by fact
qed

theorem zf_move_bij: "zf_move\<in>bij(big, zf_newbig)"
proof -
  from zf_move_inj have "zf_move\<in>bij(big, range(zf_move))" using inj_bij_range by auto
  thus ?thesis by (unfold zf_newbig_def)
qed

theorem zf_move_embed [simp]: "x\<in>small \<Longrightarrow> zf_move`(embed`x) = x"
proof -
  assume "x\<in>small"
  with embed_fun have "embed`x\<in>range(embed)" by (rule apply_rangeI)
  hence s1: "embed`x\<in>small2" by (unfold small2_def)
  with small2_sub_big have "embed`x\<in>big" by auto
  with s1 have "zf_move`(embed`x) = spec`(embed`x)" by (auto simp add: zf_move_def zf_move_fun_def)
  also from embed_inj `x\<in>small` spec_def have "spec`(embed`x) = x" by auto
  finally show "zf_move`(embed`x) = x" by auto
qed

theorem zf_embed_move: "zf_move O embed = id(small)"
proof -
  have "\<forall>y\<in>small. zf_move`(embed`y) = y" by simp
  moreover
  have "embed: small\<rightarrow>big" by (rule embed_fun)
  moreover
  have "zf_move: big\<rightarrow>zf_newbig" by (rule zf_move_fun)
  ultimately show ?thesis using comp_eq_id_iff1 by blast
qed

end

text {* Next prove that ZF generalization is an arbitrary generalization: *}
sublocale ZF_generalization \<subseteq> arbitrary_generalization small big embed zf_newbig zf_move
proof
  show "zf_move\<in>bij(big, zf_newbig)" using zf_move_bij by auto
  show "zf_move O embed = id(small)" using zf_embed_move by auto
qed

end